import os

from python_search.apps.terminal import Terminal
from python_search.config import PythonSearchConfiguration
from python_search.environment import is_mac


class FzfInKitty:
    """
    Renders the search ui using fzf + termite terminal
    """

    FONT_SIZE = 15
    PREVIEW_PERCENTAGE_SIZE = 50
    HEIGHT = 127
    WIDTH = 950
    RANK_TIE_BREAK = "begin,index"

    configuration: PythonSearchConfiguration

    def __init__(self, configuration: PythonSearchConfiguration):
        self.height = FzfInKitty.HEIGHT
        self.width = FzfInKitty.WIDTH

        self.preview_cmd = f"python_search _preview_entry {{}} "
        self.title = configuration.APPLICATION_TITLE
        self.configuration = configuration

        import logging
        import sys

        logger = logging.getLogger(name="search_ui")
        logger.addHandler(logging.StreamHandler(sys.stdout))
        logger.setLevel(logging.DEBUG)

        self._logger = logger

    def run(self) -> None:
        self._launch_terminal(self._fzf_cmd())

    def _fzf_cmd(self):
        FZF_LIGHT_THEME = "fg:#4d4d4c,bg:#ffffff,hl:#d7005f,info:#4271ae,prompt:#8959a8,pointer:#d7005f,marker:#4271ae,spinner:#4271ae,header:#4271ae,fg+:#4d4d4c,bg+:#ffffff,hl+:#d7005f"
        THEME = f"--color={FZF_LIGHT_THEME}"  # for more fzf options see: https://www.mankier.com/1/fzf#
        cmd = f"""bash -c ' export SHELL=bash ; {self._get_rankging_generate_cmd()} | \
        fzf \
        --tiebreak={FzfInKitty.RANK_TIE_BREAK} \
        --extended \
        --reverse \
        --info=inline \
        --cycle \
        --no-hscroll \
        --hscroll-off=0 \
        --preview "{self.preview_cmd}" \
        --preview-window=right,{FzfInKitty.PREVIEW_PERCENTAGE_SIZE}%,wrap,border-left \
        -i \
        --border=none \
        --margin=0% \
        --padding=0% \
        {self._run_key("enter")} \
        {self._run_key("double-click")} \
        {self._edit_key('ctrl-e')} \
        {self._edit_key('right-click')} \
        --bind "alt-enter:execute-silent:(nohup python_search run_key {{}} --query_used {{q}} & disown)" \
        --bind "alt-m:execute-silent:(nohup python_search edit_main {{}} & disown)" \
        --bind "ctrl-l:clear-query" \
        --bind "ctrl-l:+first" \
        --bind "ctrl-k:execute-silent:(python_search _copy_key_only {{}} && kill -9 $PPID)" \
        --bind "ctrl-c:execute-silent:(nohup python_search _copy_entry_content {{}} && kill -9 $PPID)" \
        --bind "ctrl-s:execute-silent:(nohup python_search search_edit {{}} && kill -9 $PPID)" \
        --bind "ctrl-r:reload:({self._get_rankging_generate_cmd(reload=True)})" \
        --bind "ctrl-t:execute-silent:(notify-send testjean)" \
        --bind "ctrl-f:first" \
        --bind "shift-up:first" \
        --bind "esc:abort" \
        --bind "ctrl-d:abort"  \
        {THEME}  ; exit 0
        '
        """

        self._logger.info("Cmd: ", cmd)
        return cmd

    def _run_key(self, shortcut) -> str:
        return f"""--bind "{shortcut}:execute-silent:(LOG_FILE=/tmp/log_run_key_fzf nohup python_search run_key {{}} --query_used {{q}} && kill -9 $PPID)"  """

    def _edit_key(self, shortcut) -> str:
        return f"""--bind "{shortcut}:execute-silent:(nohup python_search edit_key {{}} & disown && kill -9 $PPID ) "  """

    def _get_rankging_generate_cmd(self, reload=False):
        # in mac we need tensorflow to be installed via conda
        if self.configuration.supported_features.is_dynamic_ranking_supported():
            if reload:
                return f"curl -s localhost:8000/ranking/reload_and_generate"

            return f"curl -s localhost:8000/ranking/generate"

        return f"python_search ranking generate"

    def _launch_terminal(self, internal_cmd: str) -> None:

        font = "FontAwesome"
        if is_mac():
            font = "Pragmata Pro"

        launch_cmd = f"""nice -19 kitty \
        --title {self.title} \
        -o draw_minimal_borders=yes \
        -o placement_strategy=top-left \
        -o window_border_width=0 \
        -o window_padding_width=0 \
        -o hide_window_decorations=titlebar-only \
        -o background_opacity=1 \
        -o tab_bar_style=hidden  \
        -o initial_window_width={self.width}  \
        -o initial_window_height={self.height} \
        -o font_family="{font}" \
        -o font_size={FzfInKitty.FONT_SIZE} \
        {Terminal.GLOBAL_TERMINAL_PARAMS} \
         {internal_cmd}
        """
        self._logger.info(f"Command performed:\n {internal_cmd}")
        result = os.system(launch_cmd)
        if result != 0:
            raise Exception("Search run fzf projection failed")


if __name__ == "__main__":
    import fire

    fire.Fire()



import datetime
import logging
import sys

from python_search.config import ConfigurationLoader
from python_search.search.entries_loader import EntriesLoader
from python_search.search.next_item_predictor.features.entry_embeddings.entry_embeddings import (
    RedisEmbeddingsReader,
)
from python_search.search.next_item_predictor.inference.inference import Inference


class Evaluate:
    """
    Central place to evaluate the quality of the _model
    """

    def __init__(self):
        self.month = datetime.datetime.now().month
        self.NUM_OF_TOP_RESULTS = 9
        self.NUM_OF_BOTTOM_RESULTS = 4
        logging.basicConfig(
            level=logging.INFO, handlers=[logging.StreamHandler(sys.stdout)]
        )

        self.configuration = ConfigurationLoader().load()

    def evaluate(self, run_id=None):
        """
        Evaluate a _model against our search
        """
        logging.info("Evaluate _model")
        print("Run id ", run_id)
        self.all_latest_keys = EntriesLoader.load_all_keys()
        self.embeddings_keys_latest = RedisEmbeddingsReader().load(self.all_latest_keys)

        keys_to_test = [
            # "my beat81 bookings",
            "set current project as reco",
            "days quality tracking life good day",
        ]

        inference = Inference(configuration=self.configuration, run_id=run_id)

        for key in keys_to_test:
            result = inference.get_ranking(forced_previous_key=key)
            print(f"Key: {key}")

            print(f"Top")
            for i in result[0 : self.NUM_OF_TOP_RESULTS]:
                print(f"    {i}")

            print(f"Bottom")
            for i in result[-self.NUM_OF_BOTTOM_RESULTS :]:
                print(f"    {i}")



import os
from typing import Any, Tuple

import mlflow
import numpy as np
from keras import layers
from keras.models import Sequential

Model = Any


class TrainKeras:

    EPOCHS = 30
    TEST_SPLIT_SIZE = 0.10
    BATCH_SIZE = 128

    def __init__(self, epochs=None):
        if not epochs:
            epochs = TrainKeras.EPOCHS
        self.epochs = epochs
        # enable the profiling scaffolding
        os.environ["TIME_IT"] = "1"

    def train(self, X_train, X_test, Y_train, Y_test) -> Model:

        print("Starting train with N epochs, N=", self.epochs)
        print(
            {
                "shape_x_train": X_train.shape,
                "shape_x_test": X_test.shape,
                "shape_y_train": Y_train.shape,
                "shape_y_test": Y_test.shape,
                "X_train has nan: ": np.any(np.isnan(X_train)),
                "Y_train has nan: ": np.any(np.isnan(Y_train)),
                "X_test has nan: ": np.any(np.isnan(X_test)),
                "y_test has nan: ": np.any(np.isnan(Y_test)),
            }
        )

        model = Sequential()
        model.add(layers.Dense(128, activation="relu"))
        model.add(layers.Dropout(0.5))
        model.add(layers.Dense(64, activation="relu"))
        model.add(layers.Dropout(0.5))
        model.add(layers.Dense(1))
        model.compile(optimizer="rmsprop", loss="mse", metrics=["mae", "mse"])

        model.fit(
            X_train,
            Y_train,
            epochs=self.epochs,
            batch_size=TrainKeras.BATCH_SIZE,
            validation_data=(X_test, Y_test),
        )
        # self._plot_training_history(history)

        return model

    def _plot_training_history(self, history):
        import matplotlib.pyplot as plt

        loss = history.history["loss"]
        val_loss = history.history["val_loss"]

        epochs_range = range(1, self.epochs + 1)
        plt.plot(epochs_range, loss, "bo", label="Training Loss")
        plt.plot(epochs_range, val_loss, "b", label="Validation Loss")
        plt.title("Training and validation loss")
        plt.xlabel("Epochs")
        plt.ylabel("Loss")
        plt.legend()
        plt.show()



from typing import Optional

import mlflow

from python_search.config import DataConfig


def configure_mlflow(experiment_name: Optional[str] = None):
    """
    setts up an mlflow uri and experiment name
    """

    if not experiment_name:
        experiment_name = DataConfig.NEXT_ITEM_EXPERIMENT_NAME

    data = {
        "experiment_name": experiment_name,
        "tracking_uri": f"file:{DataConfig.MLFLOW_MODELS_PATH}",
    }

    print("MLFlow configuration:", data)

    mlflow.set_tracking_uri(data["tracking_uri"])
    mlflow.set_experiment(data["experiment_name"])

    return mlflow



from __future__ import annotations

import datetime
import json
import logging
import os

from dateutil import parser

from python_search.acronyms import generate_acronyms
from python_search.infrastructure.performance import timeit
from python_search.search.ranked_entries import RankedEntries


class FzfOptimizedSearchResults:
    """
    Builds the list of results ready to be consumed by fzf
    """

    def __init__(self):
        self._today = datetime.datetime.now()

    @timeit
    def build_entries_result(
        self, entries: RankedEntries.type, ranking_uuid: str
    ) -> str:
        """Print results"""
        position = 1
        result = ""
        for name, content in entries:
            try:
                if "snippet" in content:
                    content["snippet"] = sanitize(content["snippet"])
                if "cmd" in content:
                    content["cmd"] = sanitize(content["cmd"])

                content["key_name"] = name
                content["position"] = position
                content["generated_acronyms"] = generate_acronyms(name)
                content["uuid"] = ranking_uuid
                content["tags"] = content["tags"] if "tags" in content else []
                if "created_at" in content:
                    date_created = parser.parse(content["created_at"])
                    days_ago = (self._today - date_created).days

                    content["tags"].append(f"created_{days_ago}_days_ago")

                    if days_ago == 0:
                        content["tags"].append(f"today_created")
                    elif days_ago == 1:
                        content["tags"].append(f"yesterday_created")
                    if days_ago < 7:
                        content["tags"].append(f"this_week_created")
                    if days_ago > 7 and days_ago < 14:
                        content["tags"].append(f"previous_week_created")
                    if days_ago < 30:
                        content["tags"].append(f"this_month_created")
                    if days_ago > 30 and days_ago < 60:
                        content["tags"].append(f"previous_month_created")
                    if days_ago < 365:
                        content["tags"].append(f"this_year_created")
                    if days_ago > 365:
                        content["tags"].append(f"previous_year_created")

                content_str = json.dumps(content, default=tuple, ensure_ascii=True)
            except BaseException as e:
                logging.info(e)
                content_str = str(content)

            position = position + 1

            content_str = f"{name}:" + content_str
            #  replaces all single quotes for double ones
            #  otherwise the json does not get rendered
            content_str = content_str.replace("'", '"')
            if os.getenv("ENABLE_TIME_IT"):
                # do not print if enable timeit is on
                continue
            result += content_str + "\n"
        return result


def sanitize(content):
    result = ""
    for char in content:
        if char.isalnum():
            result += char
        else:
            result += " "

    return result



import sys
from typing import Union
from python_search.environment import is_mac


def chomp(x):
    """remove special chars from end of string"""
    if x.endswith("\r\n"):
        return x[:-2]
    if x.endswith("\n") or x.endswith("\r"):
        return x[:-1]
    return x


class Clipboard:
    def get_content(self, source="--primary") -> str:
        """
        Accepted values are --primary and --clipboard
        """

        cmd = f"xsel {source} --output"
        if is_mac():
            cmd = "pbpaste"

        import subprocess

        result = subprocess.getoutput(cmd)

        result = chomp(result)

        return result

    def set_content(self, content: Union[str, None] = None, enable_notifications=True):
        """
        Put a string in the clipboard.
        If no string is provided it tries to fetch it from stdin
        :param content:
        :param enable_notifications:
        :return:
        """

        if not content and has_stdin():
            content = sys.stdin.read()

        if not content:
            raise Exception("Tytrin to set empty to clipboard")

        def shellquote(s):
            return "'" + s.replace("'", "'\\''") + "'"

        sanitized = shellquote(content)

        clipboard_cmd = "xsel --clipboard --primary --input"
        if is_mac():
            clipboard_cmd = "pbcopy"

        cmd = f"echo {sanitized} | {clipboard_cmd}"

        if enable_notifications:
            from python_search.apps.notification_ui import send_notification

            send_notification(f"Content copied: {sanitized}")

        import os

        print(cmd)
        return os.system(cmd)


def has_stdin() -> bool:
    import select
    import sys

    return select.select(
        [
            sys.stdin,
        ],
        [],
        [],
        0.0,
    )[0]


def main():
    import fire

    fire.Fire(Clipboard)


if __name__ == "__main__":
    main()



import math


def label_formula(row):
    return (
        math.log(row["times_3"])
        + math.log(row["times_2"] * 0.5)
        + math.log(row["global_pair"] * 0.01)
        + math.log(row["times_used_previous"] * 0.001)
        + math.log(row["times_used_previous_previous"] * 0.001)
    )
